"""
Distributed under the terms of the BSD 3-Clause License.

The full license is in the file LICENSE, distributed with this software.

Author: Jun Zhu
"""
from abc import ABC
from typing import Any, Callable, Optional, Union

import redis

from .serializer import create_serializer, create_deserializer
from .schema_registry import CachedSchemaRegistry


class BaseRedisClient(ABC):
    def __init__(self, host: str, port: int, password: Optional[str] = None, *,
                 timeout: Optional[int] = None):
        """Initialization.

        :param host: hostname of the Redis server.
        :param port: port of the Redis server.
        :param password: Redis password.
        :param timeout: subscribe timeout in seconds.
        """
        self._db = redis.Redis(host=host, port=port, password=password)

        if timeout is None:
            self._timeout = timeout
        else:
            self._timeout = int(timeout * 1000)  # to milliseconds

    @property
    def client(self):
        return self._db

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self._db.close()


class RedisConsumer(BaseRedisClient):
    """Provide API for consuming data stored in Redis."""

    def __init__(self,
                 *args,
                 deserializer: Union[str, Callable] = "avro",
                 **kwargs):
        """Initialization.

        :param deserializer: deserializer type or a callable object which
            deserializes the data.
        """
        super().__init__(*args, **kwargs)
        self._stream = {}

        if callable(deserializer):
            self._unpack = deserializer
        else:
            self._unpack = create_deserializer(deserializer)

        self._schema_registry = CachedSchemaRegistry(self._db)

    def subscribe(self, stream: str) -> None:
        """Subscribe to a given stream.

        :param stream: stream name.
        """
        self._stream[stream] = '$'

    def consume(self, count: int = 1) -> ([dict], dict):
        """Consume a list of data items.

        :param count: the maximum number of data items too return.

        :raises: TimeoutError, RuntimeError
        """
        # the returned data is a list with at most 'count' items
        data = self._db.xread(
            self._stream, count, block=self._timeout)
        if not data:
            raise TimeoutError

        schema = self._schema_registry.get(data[0][0].decode())
        if schema is None:
            raise RuntimeError(
                f"Unable to retrieve schema for '{self._stream}'")

        ret = []
        for _, item in data[0][1]:
            ret.append({
                field["name"]: self._unpack(item[field["name"].encode()])
                for field in schema["fields"]
            })
        return ret, schema


class RedisProducer(BaseRedisClient):
    """Provide API for writing data into Redis."""

    def __init__(self,
                 *args,
                 serializer: Union[str, Callable] = "avro",
                 maxlen: int = 10,
                 **kwargs):
        """Initialization.

        :param serializer: serializer type or a callable object which
            serializes the data.
        :param maxlen: maximum size of the Redis stream.
        """
        super().__init__(*args, **kwargs)

        self._maxlen = maxlen

        if callable(serializer):
            self._pack = serializer
        else:
            self._pack = create_serializer(serializer)

        self._schema_registry = CachedSchemaRegistry(self._db)

    def _encode_with_schema(self, item, schema):
        return {field["name"]: self._pack(item[field["name"]])
                for field in schema["fields"]}

    def produce(self, stream: str, item: Any, schema: dict) -> str:
        """Produce data item to stream.

        :param stream: stream to produce data item to.
        :param item: data item.
        :param schema: data item schema.

        :raises: RuntimeError
        """
        stream_id = self._db.xadd(
            stream, self._encode_with_schema(item, schema),
            maxlen=self._maxlen
        )
        self._schema_registry.set(stream, schema)
        return stream_id.decode()


class RedisClient:
    def __init__(self):
        self._db = None

    def __get__(self, instance, instance_type):
        if self._db is None:
            self._db = redis.Redis(**instance.config)
        return self._db


class RedisSubscriber:
    def __init__(self):
        self._sub = None

    def __get__(self, instance, instance_type):
        if self._sub is None:
            self._sub = redis.Redis(**instance.config).pubsub()
        return self._sub
